package net.knarcraft.questmobspawns;

import me.pikamug.quests.Quests;
import me.pikamug.quests.player.Quester;
import me.pikamug.quests.quests.Quest;
import me.pikamug.quests.quests.components.Stage;
import net.knarcraft.questmobspawns.listener.MobBlockDamageListener;
import net.knarcraft.questmobspawns.listener.MobDeathListener;
import net.knarcraft.questmobspawns.util.QuestMobHelper;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Sound;
import org.bukkit.SoundCategory;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.LivingEntity;
import org.bukkit.event.entity.CreatureSpawnEvent;
import org.bukkit.plugin.java.JavaPlugin;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

/**
 * This plugin's main class
 */
public final class QuestMobSpawns extends JavaPlugin {

    private final Random random = new Random();
    private static QuestMobSpawns plugin;
    private Quests questsAPI;

    public static QuestMobSpawns getPlugin() {
        return plugin;
    }

    @Override
    public void onEnable() {
        plugin = this;
        // Set up the quests API
        questsAPI = (Quests) Bukkit.getServer().getPluginManager().getPlugin("Quests");
        if (questsAPI == null) {
            this.setEnabled(false);
            return;
        }

        // Register listeners
        Bukkit.getPluginManager().registerEvents(new MobDeathListener(), this);
        Bukkit.getPluginManager().registerEvents(new MobBlockDamageListener(questsAPI), this);

        Bukkit.getScheduler().scheduleSyncRepeatingTask(this, this::spawnMobs, 200, 30 * 20);
    }

    /**
     * Spawns mobs for all kill areas as necessary
     */
    private void spawnMobs() {
        // Find the players taking each quest
        Set<Quest> questsWithPlayers = new HashSet<>();
        Map<Stage, Set<Quester>> stagePlayerMap = new HashMap<>();
        Collection<Quester> onlineQuestTakers = questsAPI.getOnlineQuesters();

        // If no quest takers are online, abort
        if (onlineQuestTakers == null || onlineQuestTakers.isEmpty()) {
            return;
        }

        for (Quester questPlayer : onlineQuestTakers) {
            for (Quest quest : questPlayer.getCurrentQuests().keySet()) {
                questsWithPlayers.add(quest);
                Stage playerStage = questPlayer.getCurrentStage(quest);
                for (int i = 0; i < playerStage.getMobsToKill().size(); i++) {
                    // If the player has not killed enough mobs, spawn the mobs
                    if (questPlayer.getQuestDataOrDefault(quest).getMobNumKilled().get(i) >= playerStage.getMobNumToKill().get(i)) {
                        continue;
                    }

                    if (!stagePlayerMap.containsKey(playerStage)) {
                        stagePlayerMap.put(playerStage, new HashSet<>());
                    }
                    stagePlayerMap.get(playerStage).add(questPlayer);
                    break;
                }
            }
        }

        // If no players are online, abort
        if (stagePlayerMap.isEmpty()) {
            return;
        }

        for (Quest quest : questsAPI.getLoadedQuests()) {
            // Ignore quests with no players
            if (!questsWithPlayers.contains(quest)) {
                continue;
            }

            for (Stage stage : quest.getStages()) {
                for (int i = 0; i < stage.getLocationsToKillWithin().size(); i++) {
                    spawnMobsForStage(stagePlayerMap, quest, stage, i);
                }
            }
        }
    }

    /**
     * Spawns necessary mobs for a given stage
     *
     * @param questPlayerMap <p>The set of quest-takers on the given stage</p>
     * @param quest          <p>The quest to spawn mobs for</p>
     * @param stage          <p>The stage to spawn mobs for</p>
     * @param index          <p>The index of the kill location to spawn mobs for</p>
     */
    private void spawnMobsForStage(@NotNull Map<Stage, @NotNull Set<Quester>> questPlayerMap, @NotNull Quest quest,
                                   @NotNull Stage stage, int index) {
        Location killLocation = ((Location) stage.getLocationsToKillWithin().get(index)).clone();
        int radius = stage.getRadiiToKillWithin().get(index);

        // TODO: De-spawn custom mobs if players are no longer in the area (should probably save the state)

        // If no players are at this stage, abort
        if (questPlayerMap.get(stage) == null) {
            return;
        }

        // Calculate the number of mobs to spawn at once
        int maxMobs = radius / 2;
        int mobsToSpawn = calculateMobNumber(questPlayerMap, quest, stage, index, killLocation, radius, maxMobs);
        if (mobsToSpawn <= 0 || killLocation.getWorld() == null) {
            return;
        }

        // If there are a bunch of mobs in the area, don't spawn more
        EntityType entityType = (EntityType) stage.getMobsToKill().get(index);
        int nearby = entitiesInKillZone(killLocation, radius, entityType);
        if (nearby >= maxMobs) {
            return;
        }

        // Spawn entities
        for (int counter = 0; counter < mobsToSpawn; counter++) {
            int x = (int) Math.round((random.nextInt(radius * 2) - radius) + killLocation.getX());
            int z = (int) Math.round((random.nextInt(radius * 2) - radius) + killLocation.getZ());
            Block spawnBlock = killLocation.getWorld().getHighestBlockAt(x, z).getRelative(BlockFace.UP);
            Entity spawnedEntity = killLocation.getWorld().spawnEntity(spawnBlock.getLocation(), entityType);

            // Treat the spawned entity as the result of a spawner for plugins like mcMMO
            if (spawnedEntity instanceof LivingEntity) {
                CreatureSpawnEvent spawnEvent = new CreatureSpawnEvent((LivingEntity) spawnedEntity,
                        CreatureSpawnEvent.SpawnReason.SPAWNER);
                Bukkit.getPluginManager().callEvent(spawnEvent);
            }

            QuestMobHelper.setSpawnedMob(spawnedEntity);
            QuestMobHelper.setQuestData(spawnedEntity, quest.getId(), quest.getStages().indexOf(stage), index,
                    new int[]{killLocation.getBlockX(), killLocation.getBlockY(), killLocation.getBlockZ(), radius});
            // Play a sound so players won't be too surprised
            killLocation.getWorld().playSound(spawnBlock.getLocation(), Sound.ITEM_GOAT_HORN_SOUND_2,
                    SoundCategory.HOSTILE, 1F, 1F);
        }
    }

    /**
     * Counts the number of entities of the necessary type already in the kill zone
     *
     * @param killLocation <p>The location of the kill location's centre</p>
     * @param radius       <p>The radius of the kill zone</p>
     * @param entityType   <p>The type of entity to kill in the kill zone</p>
     * @return <p></p>
     */
    private int entitiesInKillZone(@NotNull Location killLocation, int radius, @NotNull EntityType entityType) {
        if (killLocation.getWorld() == null) {
            return 0;
        }
        Collection<Entity> nearby = killLocation.getWorld().getNearbyEntities(killLocation, radius, 10, radius,
                (entity -> entity.getType() == entityType));
        List<Entity> tooFar = new ArrayList<>();
        for (Entity entity : nearby) {
            Location entityLocation = entity.getLocation().clone();
            entityLocation.setY(0);
            if (QuestMobHelper.get2dDistance(entityLocation, killLocation) > radius) {
                tooFar.add(entity);
            }
        }
        nearby.removeAll(tooFar);
        return nearby.size();
    }

    /**
     * Calculates the number of mobs that should be spawned at once
     *
     * @param questPlayerMap <p>The map for which players are on which stage</p>
     * @param quest          <p>The quest a mob is spawned for</p>
     * @param stage          <p>The stage a mob is spawned for</p>
     * @param index          <p>The index of the task to spawn a mob for</p>
     * @param killLocation   <p>The centre of the kill location</p>
     * @param radius         <p>The radius of the kill location</p>
     * @param maxMobs        <p>The maximum amount of mobs allowed in the kill zone</p>
     * @return <p>The number of mobs to spawn</p>
     */
    private int calculateMobNumber(@NotNull Map<Stage, @NotNull Set<Quester>> questPlayerMap, @NotNull Quest quest,
                                   @NotNull Stage stage, int index, @NotNull Location killLocation, int radius,
                                   int maxMobs) {
        int mobsToSpawn = 0;
        for (Quester player : questPlayerMap.get(stage)) {
            Location playerLocation = player.getPlayer().getLocation().clone();
            playerLocation.setY(0);
            if (player.getQuestDataOrDefault(quest).getMobNumKilled().get(index) < stage.getMobNumToKill().get(index) &&
                    QuestMobHelper.get2dDistance(killLocation, playerLocation) < radius) {
                mobsToSpawn++;
            }
        }
        if (mobsToSpawn < 1 || killLocation.getWorld() == null) {
            return 0;
        }
        mobsToSpawn *= radius / 5.0;
        mobsToSpawn = Math.max(mobsToSpawn, 1);
        mobsToSpawn = Math.min(maxMobs, mobsToSpawn);
        return mobsToSpawn;
    }

}
