package net.knarcraft.questmobspawns.util;

import net.knarcraft.questmobspawns.QuestMobSpawns;
import org.bukkit.Location;
import org.bukkit.NamespacedKey;
import org.bukkit.entity.Entity;
import org.bukkit.persistence.PersistentDataContainer;
import org.bukkit.persistence.PersistentDataType;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * A helper class for quest mobs
 */
public class QuestMobHelper {

    private static final NamespacedKey questMobKey = new NamespacedKey(QuestMobSpawns.getPlugin(), "questMob");
    private static final NamespacedKey questIdKey = new NamespacedKey(QuestMobSpawns.getPlugin(), "questId");
    private static final NamespacedKey stageIdKey = new NamespacedKey(QuestMobSpawns.getPlugin(), "stageId");
    private static final NamespacedKey mobIndexKey = new NamespacedKey(QuestMobSpawns.getPlugin(), "mobIndex");
    private static final NamespacedKey killZoneLocationKey = new NamespacedKey(QuestMobSpawns.getPlugin(), "killZoneLocationKey");

    /**
     * Checks whether the given entity has been spawned by this plugin
     *
     * @param entity <p>The entity to check</p>
     * @return <p>True if the entity was spawned by this plugin</p>
     */
    public static boolean isSpawnedMob(@NotNull Entity entity) {
        return Boolean.TRUE.equals(entity.getPersistentDataContainer().get(questMobKey, PersistentDataType.BOOLEAN));
    }

    /**
     * Marks the given entity as spawned by this plugin
     *
     * @param entity <p>The entity to mark</p>
     */
    public static void setSpawnedMob(@NotNull Entity entity) {
        entity.getPersistentDataContainer().set(questMobKey, PersistentDataType.BOOLEAN, true);
    }

    /**
     * Gets the id of the quest the given entity was spawned for
     *
     * @param entity <p>The entity to get the quest id from</p>
     * @return <p>The quest id, or null if not a quest mob</p>
     */
    public static @Nullable String getQuestId(@NotNull Entity entity) {
        return entity.getPersistentDataContainer().get(questIdKey, PersistentDataType.STRING);
    }

    /**
     * Gets the id of the stage the given entity was spawned for
     *
     * <p>The stage id is relative to the quest the stage belongs to.</p>
     *
     * @param entity <p>The entity to get the stage id for</p>
     * @return <p>The stage id, or null if not a quest mob</p>
     */
    public static @Nullable Integer getStageId(Entity entity) {
        return entity.getPersistentDataContainer().get(stageIdKey, PersistentDataType.INTEGER);
    }

    /**
     * Gets the index of this entity in the quest's "mobs to kill" array
     *
     * <p>This property is necessary in order to check whether the player has already reached the required number of
     * mob kills. This matters if the stage has some other clear property other than killing this type of mob.</p>
     *
     * @param entity <p>The entity to check</p>
     * @return <p>The entity's index, or null if not a quest mob</p>
     */
    public static @Nullable Integer getIndex(@NotNull Entity entity) {
        return entity.getPersistentDataContainer().get(mobIndexKey, PersistentDataType.INTEGER);
    }

    /**
     * Gets information about the mob's kill zone
     *
     * @param entity <p>The entity to get data for</p>
     * @return <p>An array containing x,y,z,radius</p>
     */
    public static int[] getKillZoneLocation(@NotNull Entity entity) {
        return entity.getPersistentDataContainer().get(killZoneLocationKey, PersistentDataType.INTEGER_ARRAY);
    }

    /**
     * Stores quest info to a mob
     *
     * @param entity           <p>The entity to store quest data to</p>
     * @param questId          <p>The id of the quest the mob was spawned for</p>
     * @param killLocationInfo <p>Information about the kill location (x, y, z, range)</p>
     */
    public static void setQuestData(@NotNull Entity entity, @NotNull String questId, int stageId, int mobIndex,
                                    int[] killLocationInfo) {
        PersistentDataContainer persistentDataContainer = entity.getPersistentDataContainer();
        persistentDataContainer.set(questIdKey, PersistentDataType.STRING, questId);
        persistentDataContainer.set(stageIdKey, PersistentDataType.INTEGER, stageId);
        persistentDataContainer.set(mobIndexKey, PersistentDataType.INTEGER, mobIndex);
        persistentDataContainer.set(killZoneLocationKey, PersistentDataType.INTEGER_ARRAY, killLocationInfo);
    }

    /**
     * Gets the 2d (x,z) distance between two locations
     *
     * @param location1 <p>The first location</p>
     * @param location2 <p>The second location</p>
     * @return <p>The 2d distance between the locations</p>
     */
    public static double get2dDistance(@NotNull Location location1, @NotNull Location location2) {
        return Math.sqrt(Math.pow(location2.getX() - location1.getX(), 2) +
                Math.pow(location2.getZ() - location1.getZ(), 2));
    }

}
